
#include <windows.h>
#include <windowsx.h>
#include <shellapi.h>
#include <htmlhelp.h>

#include "resource.h"

#include "MD5.h"

#include "mainwin.h"



#define VERSION_MAJOR   1
#define VERSION_MINOR   0


#define DEF_WIN_WIDTH   480
#define DEF_WIN_HEIGHT  120

#define DEF_STRBUF_SIZE 8192

#define TEXT_BORDER_SIZE    4
#define TEXT_LINE_GAP       4


#define CLASS_NAME  TEXT("MAINWIN")


// forwards

LRESULT CALLBACK MainWin_WinProc(HWND, UINT, WPARAM, LPARAM);

BOOL MainWin_OnCreate(HWND, LPCREATESTRUCT);
void MainWin_OnDestroy(HWND);
void MainWin_OnPaint(HWND);
void MainWin_OnDropFiles(HWND, HDROP);


LRESULT MainWin_OnMenuExec(HWND);
LRESULT MainWin_OnMenuExit(HWND);
LRESULT MainWin_OnMenuAbout(HWND);
LRESULT MainWin_OnMenuCopy(HWND);
LRESULT MainWin_OnMenuHelp(HWND);


DWORD CalcFileMD5(TCHAR*, BYTE*);


void MainWin_ShowMessage(TCHAR* szMess);


//////////////////////////////////////////////////////////////////////////////


// creation and destruction


BOOL WINAPI MainWin_Init
  (HINSTANCE hInst) 
{
  WNDCLASS wclass;

  wclass.style         = CS_HREDRAW | CS_VREDRAW;
  wclass.lpfnWndProc   = MainWin_WinProc;
  wclass.cbClsExtra    = 0;
  wclass.cbWndExtra    = 0;
  wclass.hInstance     = hInst;
  wclass.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
  wclass.hCursor       = LoadCursor(NULL, IDC_ARROW);
  wclass.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
  wclass.lpszMenuName  = NULL;
  wclass.lpszClassName = CLASS_NAME;

  return (RegisterClass(&wclass) != 0);
}


//////////////////////////////////////////////////////////////////////////////


PMAINWINCTX WINAPI MainWin_Create
  (HINSTANCE hInst) 
{
  HWND        hWnd;
  PMAINWINCTX pNewCtx;
  HMENU       hMenu;
  TCHAR       szBuf[128];

  // be smart, load the menu manually

  hMenu = LoadMenu(hInst, MAKEINTRESOURCE(IDR_MENU_MAINWIN));

  LoadString(hInst, IDS_MAINWIN_CAPTION, szBuf, sizeof(szBuf));

  hWnd = CreateWindow(CLASS_NAME,
                      szBuf,
                      WS_OVERLAPPEDWINDOW,
                      (GetSystemMetrics(SM_CXSCREEN) - DEF_WIN_WIDTH) >> 1,
                      (GetSystemMetrics(SM_CYSCREEN) - DEF_WIN_HEIGHT) >> 1,
                      DEF_WIN_WIDTH,
                      DEF_WIN_HEIGHT,
                      NULL,
                      NULL,
                      hInst,
                      NULL);
  
  if (NULL == hWnd) 
  {
    return NULL;
  }
  else 
  {
    if (NULL != (pNewCtx = (PMAINWINCTX) malloc(sizeof(MAINWINCTX)))) 
	{
      pNewCtx->hInst = hInst;
      pNewCtx->hWnd  = hWnd;

      SetWindowLong(hWnd,
                    GWL_USERDATA,
                    (LONG)pNewCtx);

      DragAcceptFiles(hWnd, TRUE);

      SetMenu(hWnd, hMenu);
    }

    // init. the context data
    pNewCtx->szText[0] = (TCHAR)'\0';
  
    return pNewCtx;
  }

}

//////////////////////////////////////////////////////////////////////////////

BOOL WINAPI MainWin_Destroy
  (PMAINWINCTX pCtx) 
{
  HWND hWnd = pCtx->hWnd;

  free(pCtx);

  return DestroyWindow(hWnd);
}

//////////////////////////////////////////////////////////////////////////////

// support

HWND WINAPI MainWin_GetHandle(PMAINWINCTX pCtx) 
{
  return pCtx->hWnd;
}

//////////////////////////////////////////////////////////////////////////////

PMAINWINCTX WINAPI MainWin_GetContext
  (HWND hWnd) 
{
  return (PMAINWINCTX) GetWindowLong(hWnd, GWL_USERDATA);
}

//////////////////////////////////////////////////////////////////////////////

void WINAPI MainWin_SetText
  (PMAINWINCTX pCtx, 
   TCHAR* szText) 
{
  lstrcpyn(pCtx->szText, szText, MAINWIN_MAX_TEXT_LEN + 1);

  EnableMenuItem(GetMenu(pCtx->hWnd), 
                 IDM_COPY, 
                 lstrlen(szText) ? MF_ENABLED : MF_GRAYED);
}

//////////////////////////////////////////////////////////////////////////////

TCHAR* WINAPI MainWin_GetText
  (PMAINWINCTX pCtx) 
{
  return pCtx->szText;
}

//////////////////////////////////////////////////////////////////////////////

// winproc

LRESULT CALLBACK MainWin_WinProc
  (HWND hWnd,
   UINT unMessage,
   WPARAM wParam,
   LPARAM lParam) 
{
  switch (unMessage) 
  {
    case WM_CREATE  : return HANDLE_WM_CREATE (hWnd, wParam, lParam, MainWin_OnCreate);
    case WM_DESTROY : return HANDLE_WM_DESTROY(hWnd, wParam, lParam, MainWin_OnDestroy);
    case WM_PAINT   : return HANDLE_WM_PAINT  (hWnd, wParam, lParam, MainWin_OnPaint);

    case WM_DROPFILES : return HANDLE_WM_DROPFILES(hWnd, wParam, lParam, MainWin_OnDropFiles);

    case WM_COMMAND :
      switch (LOWORD(wParam)) 
	  {
        case IDM_EXEC  : return MainWin_OnMenuExec (hWnd);
		case IDM_EXIT  : return MainWin_OnMenuExit (hWnd);
		case IDM_ABOUT : return MainWin_OnMenuAbout(hWnd);
		case IDM_COPY  : return MainWin_OnMenuCopy (hWnd);
		case IDM_HELP  : return MainWin_OnMenuHelp (hWnd);
	  }
  }

  return DefWindowProc(hWnd, unMessage, wParam, lParam);
}

//////////////////////////////////////////////////////////////////////////////

// message cracker

BOOL MainWin_OnCreate
  (HWND hWnd,
   LPCREATESTRUCT lpCreateStruct) 
{
  return TRUE;
}

//////////////////////////////////////////////////////////////////////////////

void MainWin_OnDestroy
  (HWND hWnd) 
{
  PostQuitMessage(0);
}

//////////////////////////////////////////////////////////////////////////////

void MainWin_OnPaint
  (HWND hWnd) 
{
  TEXTMETRIC  tm;
  PAINTSTRUCT ps;
  RECT        rect;
  HDC         hdc;
  TCHAR*      szStart;
  TCHAR*      szScan;
  TCHAR       cSave;
  int         nState;
  int         nY;
  int         nStepY;
  PMAINWINCTX pCtx;


  pCtx = MainWin_GetContext(hWnd);

  hdc = BeginPaint(hWnd, &ps);

  GetClientRect(hWnd, &rect);

  // cls
  FillRect(hdc, 
           &rect, 
           (HBRUSH)GetStockObject(WHITE_BRUSH)); 

  // print out all lines

  GetTextMetrics(hdc, &tm);
  nStepY = tm.tmHeight + tm.tmExternalLeading + TEXT_LINE_GAP;

  nState = 0;
  szStart = szScan = pCtx->szText;
  nY = TEXT_BORDER_SIZE;

  while (nState != 2) 
  {
    switch (nState) 
	{
      case 0 : 
	  {
        if (('\n' == *szScan) || ('\r' == *szScan) || ('\0' == *szScan)) 
		{
          cSave = *szScan;
          *szScan = '\0';
        
          TextOut(hdc, 
                  TEXT_BORDER_SIZE,
                  nY,
                  szStart,
                  lstrlen(szStart));

          nY += nStepY;

          *szScan = cSave;

          nState = ('\0' == *szScan) ? 2 : 1;
		}
        break;
      }

      case 1 : 
	  {
        if (('\n' != *szScan) && ('\r' != *szScan)) 
		{
          // new line detected  
          nState = 0;
          szStart = szScan;
        }
      }
    }
    szScan++;
  }

  EndPaint(hWnd, &ps);
}

//////////////////////////////////////////////////////////////////////////////

void MainWin_OnDropFiles
  (HWND hWnd, 
   HDROP hDrop) 
{
  TCHAR szFileName[MAX_PATH];

  if (DragQueryFile(hDrop, (ULONG)-1, szFileName, MAX_PATH)) 
  {
    // process the first file name only    
    DragQueryFile(hDrop, 0, szFileName, MAX_PATH);

    MainWin_ExecFileHash(MainWin_GetContext(hWnd),
                         szFileName);
  }

  DragFinish(hDrop);
}

//////////////////////////////////////////////////////////////////////////////

LRESULT MainWin_OnMenuExit
  (HWND hWnd) 
{
  SendMessage(hWnd, WM_CLOSE, 0, 0);

  return 0;
}

//////////////////////////////////////////////////////////////////////////////

LRESULT MainWin_OnMenuAbout
  (HWND hWnd) 
{
  PMAINWINCTX pCtx;
  HINSTANCE hInst;

  TCHAR szTemp[1024];
  TCHAR szBuf[1024];
  TCHAR szBuf2[128];

  pCtx = MainWin_GetContext(hWnd);

  hInst = pCtx->hInst;

  LoadString(hInst, IDS_MAINWIN_INFO, szTemp, sizeof(szTemp));
  LoadString(hInst, IDS_MAINWIN_ABOUT, szBuf2, sizeof(szBuf2));

  wsprintf(szBuf, szTemp, pCtx->szVersion);

  MessageBox(hWnd,
             szBuf,
             szBuf2,
             MB_ICONINFORMATION | MB_OK);

  return 0;
}

//////////////////////////////////////////////////////////////////////////////

void Win32ErrToCStr
  (DWORD dwErr,
   TCHAR* szBuf) 
{
  TCHAR* pTemp;

  if (FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                    NULL,
                    dwErr,
                    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                    (LPTSTR)&pTemp,
                    0,
                    NULL) == 0) 
  {
    // (this should never happen)
    lstrcpy(szBuf, TEXT("unknown error"));
  }
  else 
  {
    lstrcpy(szBuf, pTemp);
    LocalFree((HLOCAL)pTemp);
  }
}

//////////////////////////////////////////////////////////////////////////////

LRESULT MainWin_OnMenuExec
  (HWND hWnd) 
{
  OPENFILENAME ofn;
  PMAINWINCTX  pCtx;
  TCHAR        szFileName[MAX_PATH];

  szFileName[0] = '\0';

  pCtx = MainWin_GetContext(hWnd);

  ofn.lStructSize       = sizeof(OPENFILENAME);
  ofn.hwndOwner         = hWnd; 
  ofn.hInstance         = NULL; 
  ofn.lpstrFilter       = NULL; 
  ofn.lpstrCustomFilter = NULL; 
  ofn.nMaxCustFilter    = 0; 
  ofn.nFilterIndex      = 0; 
  ofn.lpstrFile         = szFileName; 
  ofn.nMaxFile          = MAX_PATH; 
  ofn.lpstrFileTitle    = NULL; 
  ofn.nMaxFileTitle     = MAX_PATH; 
  ofn.lpstrInitialDir   = pCtx->szCurrPath; 
  ofn.lpstrTitle        = NULL; 
  ofn.Flags             = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY; 
  ofn.nFileOffset       = 0; 
  ofn.nFileExtension    = 0; 
  ofn.lpstrDefExt       = NULL; 
  ofn.lCustData         = 0L; 
  ofn.lpfnHook          = NULL; 
  ofn.lpTemplateName    = NULL; 

  if (GetOpenFileName(&ofn)) 
  {
    // apparently this is possible
    lstrcpy(pCtx->szCurrPath, szFileName);

    MainWin_ExecFileHash(pCtx, szFileName);
  }

  return 0;
}

//////////////////////////////////////////////////////////////////////////////

LRESULT MainWin_OnMenuCopy
  (HWND hWnd) 
{
  HGLOBAL     hGlobal;
  LPVOID      pGlobal;
  PMAINWINCTX pCtx;

  pCtx = MainWin_GetContext(hWnd);

  // copying a text into the clipboard, it's that easy...

  hGlobal = GlobalAlloc(GHND | GMEM_SHARE, 
                        lstrlen(pCtx->szText) + sizeof(TCHAR));
  
  pGlobal = GlobalLock(hGlobal);

  lstrcpy((LPTSTR)pGlobal, pCtx->szText);

  GlobalUnlock(hGlobal);

  OpenClipboard(hWnd);  
  EmptyClipboard();

  SetClipboardData(CF_TEXT, hGlobal);

  CloseClipboard();
  
  return 0;
}

//////////////////////////////////////////////////////////////////////////////

LRESULT MainWin_OnMenuHelp
  (HWND hWnd) 
{
  PMAINWINCTX pCtx;
  TCHAR helpPath[MAX_PATH];
  TCHAR* append;
  TCHAR szCaption[256];
  TCHAR szMsg[256];
  HINSTANCE hInst;

  pCtx = MainWin_GetContext(hWnd);

  lstrcpy(append = helpPath, pCtx->szExeName);
  
  append += lstrlen(helpPath);
  
  while (append >= helpPath)
  { 
  	if ('.' == *append) break; 
  	append--;
  } 
  
  lstrcpy(append + 1, TEXT("chm"));

  if (NULL == HtmlHelp(hWnd, 
	                   helpPath,
                       HH_DISPLAY_TOPIC,
                       0))
  {
    hInst = pCtx->hInst;
  
    LoadString(hInst, IDS_MAINWIN_ERROR, szCaption, sizeof(szCaption));
    LoadString(hInst, IDS_MAINWIN_NOHELP, szMsg, sizeof(szMsg));
  
    MessageBox(hWnd,
               szMsg,
               szCaption,
               MB_ICONSTOP | MB_OK);
  }                       

  return 0;
}

//////////////////////////////////////////////////////////////////////////////

// the file checksum routine

#define IOBUF_SIZE  32768

DWORD CalcFileMD5
  (TCHAR* szFileName,
   BYTE* pDigest) 
{
  HANDLE  hFile;
  DWORD   dwRead;
  DWORD   dwErr;
  BYTE    ioBuf[IOBUF_SIZE];
  MD5CTX  hasher;
  BOOL    blReadOK;

  hFile = CreateFile(szFileName,
                     GENERIC_READ,
                     FILE_SHARE_READ,
                     NULL,
                     OPEN_EXISTING,
                     FILE_FLAG_SEQUENTIAL_SCAN,
                     0);
  
  if (INVALID_HANDLE_VALUE == hFile) 
  {
    return GetLastError();
  }

  MD5_Initialize(&hasher);

  while (TRUE == (blReadOK = ReadFile(hFile,
                                      ioBuf,
                             	      IOBUF_SIZE,
                             	      &dwRead,
                                      NULL))) 
  {
    if (0 == dwRead) break;

    MD5_Update(&hasher, ioBuf, dwRead);
  }

  if (!blReadOK) 
  {
    dwErr = GetLastError();
  
    if (dwErr != NOERROR) 
	{
      dwErr = (ERROR_INVALID_HANDLE == dwErr) ? NOERROR : dwErr;
    }
  }
  else 
  {
    dwErr = NOERROR;

    MD5_Final(pDigest, &hasher);
  }

  CloseHandle(hFile);

  return dwErr;
}

//////////////////////////////////////////////////////////////////////////////

void MainWin_ExecFileHash
  (PMAINWINCTX pCtx,
   TCHAR* szFileName) 
{
  DWORD dwErr;
  unsigned char digest[MD5_DIGESTSIZE];
  TCHAR szBuf[DEF_STRBUF_SIZE]; 
  TCHAR szTemp[DEF_STRBUF_SIZE]; 
    
  LoadString(pCtx->hInst, IDS_MAINWIN_PROGRESS, szBuf, DEF_STRBUF_SIZE);

  wsprintf(szTemp,
           szBuf,
           szFileName);
  MainWin_SetText(pCtx, szTemp);

  InvalidateRect(pCtx->hWnd, NULL, FALSE);
  MainWin_OnPaint(pCtx->hWnd);

  dwErr = CalcFileMD5(szFileName, digest) ;
  
  MainWin_SetText(pCtx, "");
  // FIXME: better to pack this into MainWin_OnPaint()? 
  InvalidateRect(pCtx->hWnd, NULL, FALSE);
  MainWin_OnPaint(pCtx->hWnd);

  if (NOERROR == dwErr) 
  {
    LoadString(pCtx->hInst, IDS_MAINWIN_REPORT, szBuf, DEF_STRBUF_SIZE);

    wsprintf(szTemp, 
             szBuf,
             szFileName,
             digest[ 0], digest[ 1], digest[ 2], digest[ 3],
             digest[ 4], digest[ 5], digest[ 6], digest[ 7],
             digest[ 8], digest[ 9], digest[10], digest[11],
             digest[12], digest[13], digest[14], digest[15]);

    MainWin_SetText(pCtx, szTemp);
  }
  else 
  {
    Win32ErrToCStr(dwErr, szBuf);

    MessageBox(pCtx->hWnd,
               szBuf,
               NULL,
               MB_ICONSTOP);
  }

  InvalidateRect(pCtx->hWnd, NULL, FALSE);
  MainWin_OnPaint(pCtx->hWnd);
}
